// Copyright (c) 2022 Semjon Geist.
#ifndef INST__CORNFLAKES_BINDINGS_H_
#define INST__CORNFLAKES_BINDINGS_H_

#include <digest.h>
#include <ini.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <complex>
#include <string>
#include <vector>

namespace py = pybind11;
#define STRINGIFY(x) #x
#define MACRO_STRINGIFY(x) STRINGIFY(x)

#endif  // INST__CORNFLAKES_BINDINGS_H_
